# -*- coding: utf-8 -*-
"""pyaddict"""
from __future__ import annotations
__copyright__ = ("Copyright (c) 2022 https://github.com/dxstiny")

from typing import Union, Optional, Any, Type, TypeVar, overload
from abc import ABC, abstractmethod


T = TypeVar("T")


class ICommon(ABC):
    """Common interface for all public class."""

    @abstractmethod
    def optionalCast(self, key: Union[str, int], type_: Type[T]) -> Optional[T]:
        """you want to handle non-existing keys, but don't care about deviating types"""

    @abstractmethod
    def optionalGet(self, key: Union[str, int], type_: Type[T]) -> Optional[T]:
        """you want to handle deviating data (extens dict.get() with a type)"""

    @abstractmethod
    def ensure(self, key: Union[str, int], type_: Type[T], default: Optional[T] = None) -> T:
        """you don't really know the data, but want to have a default value"""

    @abstractmethod
    def ensureCast(self, key: Union[str, int], type_: Type[T], default: Optional[T] = None) -> T:
        """
        you don't really know the data, don't care about the type
        but want to have a default value
        """


class IExtended(ICommon):
    """Extended interface for all public class."""

    @abstractmethod
    def assertGet(self, key: Union[str, int], type_: Type[T]) -> T:
        """you know the data"""

    @overload
    @abstractmethod
    def get(self, index: Union[str, int]) -> Optional[Any]: ...
    @overload
    @abstractmethod
    def get(self, index: Union[str, int], default: Optional[T] = None) -> Union[Any, T]: ...
    @abstractmethod
    def get(self, key: Union[str, int], default: Optional[T] = None) -> Union[T, Any]:
        """get exception safe value"""

    @abstractmethod
    def __getitem__(self, key: Union[str, int]) -> Any:
        """indexable"""



# Your list contains different types

from pyaddict import JList

array = [ 1, 2.0, "Hello", True, [ 1, 2, 3 ], { "a": 1, "b": 2 } ] # I hope you never have to use such a list
jlist = JList(array)

# with normal lists you would have to do this
print(array[0]) # 1
print(array[1]) # 2.0
print(array[2]) # "Hello"
print(array[3]) # True
print(array[4]) # [ 1, 2, 3 ]
print(array[5]) # { "a": 1, "b": 2 }
# easy until now
print(array[0] + array[1]) # 3.0
try:
    print(array[1] + array[2]) # TypeError: unsupported operand type(s) for +: 'float' and 'str'
except TypeError as e:
    print(e)
try:
    print(array[2] + array[3]) # TypeError: can only concatenate str (not "bool") to str
except TypeError as e:
    print(e)
try:
    print(array[4] + array[5]) # TypeError: can only concatenate list (not "dict") to list
except TypeError as e:
    print(e)
# but this is probably not what you want

# JList can handle this
print(jlist.ensureCast(0, int)) # 1
print(jlist.ensureCast(1, float)) # 2.0
print(jlist.ensureCast(2, str)) # "Hello"
print(jlist.ensureCast(3, bool)) # True
print(jlist.ensureCast(4, list)) # [ 1, 2, 3 ]
print(jlist.ensureCast(5, dict)) # { "a": 1, "b": 2 }
# just like normal lists, but with type safety
print(jlist.ensureCast(0, int) + jlist.ensureCast(1, int)) # 3.0
print(jlist.ensureCast(1, str) + jlist.ensureCast(2, str)) # "2.0Hello"
print(jlist.ensureCast(2, str) + jlist.ensureCast(3, str)) # "HelloTrue"
print(jlist.ensureCast(4, list) + jlist.ensureCast(5, list)) # [ 1, 2, 3, "a", "b" } ]



# You worry about IndexOutOfRange exceptions

from pyaddict import JList

gallery = [ "image.png", "otherImage.png" ]
gallery2 = [ ] # empty gallery :/

jlist = JList(gallery)
jlist2 = JList(gallery2)

# with normal lists
wallpaper = gallery[0] # "image.png"
print(wallpaper) # "image.png"
try:
    wallpaper = gallery2[0] # IndexError: list index out of range
except IndexError as e:
    print(e)

# JList can handle this
wallpaper = jlist.ensure(0, str) # "image.png"
print(wallpaper) # "image.png"
wallpaper = jlist2.ensure(0, str) # ""
print(wallpaper) # ""



from pyaddict.pyaddict import JDict, JList

adventurer = {
  "name": 'Alice',
  "cat": {
    "name": 'Dinah'
  },
  "animals": [{
    "name": 'Dinah',
    "type": 'cat'
  }, {
    "name": 'Rabbit',
    "type": 'bunny',
    "age": 2,
    "ages": "2"
  }]
}
chain = JDict(adventurer).chain()

print(chain.optionalGet("dog?.name", str)) # None
print(chain.optionalGet("pets?.[1]?.name", str)) # None

print(chain.optionalGet("animals?.[1]?.name", int)) # None
print(chain.optionalCast("animals?.[1]?.name", str)) # Rabbit
print(chain.optionalGet("animals?.[1]?.ages", int)) # None
print(chain.optionalCast("animals?.[1]?.ages", int)) # 2

print(chain.ensure("animals?.[1]?.type", int)) # 0
print(chain.ensure("animals?.[1]?.ages", int)) # 0
print(chain.ensureCast("animals?.[1]?.ages", int)) # 2

print(chain.ensure("animals?.[1]", JDict)) # {}
print(chain.ensureCast("animals?.[1]", JDict)) # {'name': 'Rabbit', 'type': 'bunny', 'age': 2, 'ages': '2'}

# jlist

animals = chain.ensureCast("animals", JList).chain()
print(animals.ensure("[1]?.name", str)) # Rabbit
print(animals.ensure("[1]?.age", int)) # 0
print(animals.ensureCast("[1]?.age", int)) # 2

print(animals["[1].name"]) # Rabbit
print(animals["[1].age"]) # 2
print(animals["[2].ages"]) # IndexError
print(animals["[2]?.ages"]) # None



from pyaddict.schema import Object, String, Integer, Array, Float

dict = {
    "name": "John",
    "age": 30,
    "cars": [
        {"model": "BMW 230", "mpg": 27.5},
        {"model": "Ford Edge", "mpg": 24.1}
    ]
}

schema = Object({
    "name": String(),
    "age": Integer(),
    "cars": Array(Object({
        "model": String(),
        "mpg": Float()
    })).min(1, False).max(2)
})

print(schema.error(dict)) # None

schema = Object({
    "name": String(),
    "age": String().coerce(),
    "dogs": Array(String()).min(1).optional()
}).withAdditionalProperties()

print(schema.error(dict)) # None

badSchema = Object({
    "name": String().min(5),
    "age": Float(),
    "cars": Object()
})

print(badSchema.error(dict)) # ValidationError(expected 4 to be greater than or equal to 5, name: min)


# -*- coding: utf-8 -*-
"""pyaddict"""
from __future__ import annotations
__copyright__ = ("Copyright (c) 2022 https://github.com/dxstiny")

import json
import re
from typing import Any, List, Optional, Tuple, Type, TypeVar, Union, overload

from pyaddict.types import JArray, JObject
from pyaddict.interfaces.common import ICommon, IExtended

T = TypeVar("T")

__all__ = ["JDict", "JList", "JListIterator"]


class JDict(JObject, IExtended):
    """dict extractor"""
    def __init__(self, data: Optional[JObject] = None) -> None:
        self._data = data or { }
        super().__init__(self._data)

    def __bool__(self) -> bool:
        return bool(self._data)

    def assertGet(self, key: str, type_: Type[T]) -> T:
        """
        asserts the key exists & is the specified type

        if the key does not exist, an exception is raised
        if the key exists but is not the specified type, an exception is raised

        raises AssertionError

        Usage: you know the key exists & is the specified type
        """
        assert key in self._data
        val = self._data[key]
        assert isinstance(val, type_)
        return val

    def optionalGet(self, key: str, type_: Type[T]) -> Optional[T]:
        """
        tries to get the key

        if the key does not exist, None is returned
        if the key exists but is not the specified type, None is returned

        Usage: you don't know if the key exists
        """
        try:
            value = self._data[key]
            if isinstance(value, type_):
                return value
            return None
        except: # pylint: disable=bare-except
            return None

    def optionalCast(self, key: str, type_: Type[T]) -> Optional[T]:
        """
        tries to get the key & cast to the specified type

        if the key does not exist, None is returned

        Usage: you don't know if the key exists & you don't care if it's the specified type
        """
        try:
            return type_(self._data[key]) # type: ignore
        except: # pylint: disable=bare-except
            return None

    def ensure(self, key: str, type_: Type[T], default: Optional[T] = None) -> T:
        """
        ensures the key exists & is the specified type

        if the key does not exist, the default value is returned
        if the key exists but is not the specified type, the default value is returned

        Usage: you don't know if the key exists, you care if it's the specified type
        but want a default value
        """
        try:
            value = self._data[key]
            if isinstance(value, type_):
                return value

            return default or type_()
        except: # pylint: disable=bare-except
            return default or type_()

    def ensureCast(self, key: str, type_: Type[T], default: Optional[T] = None) -> T:
        """
        ensures the key exists & is the specified type

        if the key does not exist, the default value is returned
        if the key exists but is not the specified type, the value is cast to the specified type

        Usage: you don't know if the key exists, you don't care if it's the specified type &
        you want a default value
        """
        try:
            return type_(self._data[key]) # type: ignore
        except: # pylint: disable=bare-except
            return default or type_()

    def toString(self, indent: Optional[int] = None) -> str:
        """returns the string representation of the data"""
        return json.dumps(self._data, indent=indent)

    def toDict(self) -> JObject:
        """returns the dict representation of the data"""
        return self._data

    @staticmethod
    def fromString(string: str) -> JDict:
        """returns a JDict from a string"""
        value = json.loads(string)
        if isinstance(value, dict):
            return JDict(value)
        return JDict()

    def chain(self) -> JChain:
        """
        equivalent to js optional chaining
        """
        return JChain(jdict = self)


class JList(JArray, IExtended):
    """list extractor"""
    def __init__(self, data: Optional[JArray] = None) -> None:
        self._data = data or [ ]
        super().__init__(self._data)

    def __bool__(self) -> bool:
        return bool(self._data)

    @overload
    def get(self, index: int) -> Optional[Any]: ...
    @overload
    def get(self, index: int, default: Optional[T] = None) -> Union[Any, T]: ... # pylint: disable=arguments-differ

    def get(self, index: int, default: Optional[T] = None) -> Union[Any, T]: # pylint: disable=arguments-differ
        """
        returns the value at the specified index

        if the index does not exist, None is returned
        """
        if index < len(self._data):
            return self._data[index]
        return default

    def assertGet(self, index: int, type_: Type[T]) -> T: # pylint: disable=arguments-renamed
        """
        asserts the index exists & is the specified type

        if the index does not exist, an exception is raised
        if the index exists but is not the specified type, an exception is raised

        Usage: you know the index exists & is the specified type
        """
        assert index < len(self._data)
        val = self._data[index]
        assert isinstance(val, type_)
        return val

    def optionalGet(self, index: int, type_: Type[T]) -> Optional[T]: # pylint: disable=arguments-renamed
        """
        tries to get the index

        if the index does not exist, None is returned
        if the index exists but is not the specified type, None is returned

        Usage: you don't know if the index exists
        """
        try:
            value = self._data[index]
            if isinstance(value, type_):
                return value
            return None
        except: # pylint: disable=bare-except
            return None

    def optionalCast(self, index: int, type_: Type[T]) -> Optional[T]: # pylint: disable=arguments-renamed
        """
        tries to get the index & cast to the specified type

        if the index does not exist, None is returned

        Usage: you don't know if the index exists & you don't care if it's the specified type
        """
        try:
            return type_(self._data[index]) # type: ignore
        except: # pylint: disable=bare-except
            return None

    def ensure(self, index: int, type_: Type[T], default: Optional[T] = None) -> T: # pylint: disable=arguments-renamed
        """
        ensures the index exists & is the specified type

        if the index does not exist, the default value is returned
        if the index exists but is not the specified type, the default value is returned

        Usage: you don't know if the index exists, you care if it's the specified type
        but want a default value
        """
        try:
            value = self._data[index]
            if isinstance(value, type_):
                return value

            return default or type_()
        except: # pylint: disable=bare-except
            return default or type_()

    def ensureCast(self, index: int, type_: Type[T], default: Optional[T] = None) -> T: # pylint: disable=arguments-renamed
        """
        ensures the index exists & is the specified type

        if the index does not exist, the default value is returned
        if the index exists but is not the specified type, the value is cast to the specified type

        Usage: you don't know if the index exists, you don't care if it's the specified type &
        you want a default value
        """
        try:
            return type_(self._data[index]) # type: ignore
        except: # pylint: disable=bare-except
            return default or type_()

    def iterator(self) -> JListIterator:
        """
        returns an iterator selector for the list
        """
        return JListIterator(self)

    def toString(self, indent: Optional[int] = None) -> str:
        """returns the string representation of the data"""
        return json.dumps(self._data, indent=indent)

    def toList(self) -> JArray:
        """returns the list representation of the data"""
        return self._data

    @staticmethod
    def fromString(string: str) -> JList:
        """returns a JList from a string"""
        value = json.loads(string)
        if isinstance(value, list):
            return JList(value)
        return JList()

    def chain(self) -> JChain:
        """
        equivalent to js optional chaining
        """
        return JChain(jlist = self)


class JListIterator(JArray, ICommon):
    """list iterator"""
    def __init__(self, data: JList) -> None:
        self._data = data
        super().__init__(self._data)

    def assertGet(self, type_: Type[T]) -> List[T]: # pylint: disable=arguments-differ
        """
        iterates over the list & asserts each value is the specified type
        (which is usually the case for arrays)

        Usage: you know the values are the specified type
        """
        return [ self._data.assertGet(i, type_) for i, _ in enumerate(self._data) ]

    def optionalGet(self, type_: Type[T]) -> List[Optional[T]]: # pylint: disable=arguments-differ
        """
        iterates over the list & tries to get each with the specified type

        Usage: you don't know if the values are the specified type
        """
        return [ self._data.optionalGet(i, type_) for i, _ in enumerate(self._data) ]

    def optionalCast(self, type_: Type[T]) -> List[Optional[T]]: # pylint: disable=arguments-differ
        """
        iterates over the list & tries to cast each value to the specified type

        Usage: you don't know if the values are the specified type
        """
        return [ self._data.optionalGet(i, type_) for i, _ in enumerate(self._data) ]

    def ensure(self, type_: Type[T], default: Optional[T] = None) -> List[T]: # pylint: disable=arguments-differ
        """
        iterates over the list & ensures each value is the specified type

        Usage: you don't know if the values are the specified type & you want a default value
        """
        return [ self._data.ensure(i, type_, default) for i, _ in enumerate(self._data) ]

    def ensureCast(self, type_: Type[T], default: Optional[T] = None) -> List[T]: # pylint: disable=arguments-differ
        """
        iterates over the list & ensures each value is the specified type

        Usage: you don't know if the values are the specified type & you want a default value
        """
        return [ self._data.ensureCast(i, type_, default) for i, _ in enumerate(self._data) ]


class _ChainLink:
    """chain link"""
    def __init__(self, key: str) -> None:
        self._key: Union[str, int] = key
        self._optional = False
        self._index = False

        if key.endswith("?"):
            self._key = key[:-1]
            self._optional = True
        if key.startswith("[") and key.endswith("]"):
            self._key = int(key[1:-1])
            self._index = True

    def __repr__(self) -> str:
        return f"ChainLink({self._key}, optional={self._optional}, index={self._index})"

    @property
    def key(self) -> Union[str, int]:
        """returns the key"""
        return self._key

    @property
    def intKey(self) -> int:
        """returns the key as an int"""
        assert isinstance(self._key, int)
        return self._key

    @property
    def strKey(self) -> str:
        """returns the key as a string"""
        assert isinstance(self._key, str)
        return self._key

    @property
    def optional(self) -> bool:
        """returns true if the link is optional"""
        return self._optional

    @property
    def index(self) -> bool:
        """returns true if the link is an index"""
        return self._index


class JChain(ICommon):
    """optional chaining"""
    def __init__(self,
                 jdict: Optional[JDict] = None,
                 jlist: Optional[JList] = None) -> None:

        if jdict is None and jlist is None:
            raise Exception("JChain requires either a JDict or JList")

        self._jdict: Optional[JDict] = jdict
        self._jlist: Optional[JList] = jlist

    def _isDict(self) -> bool:
        return bool(self._jdict)

    def _isList(self) -> bool:
        return bool(self._jlist)

    def _createChainLinks(self, chain: str) -> List[_ChainLink]: # pylint: disable=no-self-use
        chain = re.sub(r"(\w)\[", r"\1.[", chain)
        return [ _ChainLink(key) for key in chain.split(".") ]

    def __getitem__(self, name: str) -> Any:
        value = self._prepare(name)
        if value is None:
            return None
        parent, key = value
        return parent.get(key)

    def _prepare(self,
                 chain: str,
                 preventException: bool = False) -> Optional[Tuple[IExtended, Union[str, int]]]:
        """
        prepares the chain for execution

        returns None if the chain is invalid
        else, returns (parent, last_key)
        """
        links = self._createChainLinks(chain)
        last = links.pop()
        nullable: Optional[Union[JObject, JArray]] = self._jdict if self._isDict() else self._jlist

        assert nullable is not None

        value = nullable

        for link in links:
            if preventException or link.optional:
                if link.index and link.intKey >= len(list(value)): # jlist
                    return None # couldn't resolve
                if not link.index and link.strKey not in value: # jdict
                    return None # couldn't resolve
            if isinstance(value, (JDict, dict)):
                value = value[link.strKey]
                continue
            assert isinstance(value, (JList, list))
            value = value[link.intKey]

        if last.index and isinstance(value, list):
            return JList(value), last.key
        if not last.index and isinstance(value, dict):
            return JDict(value), last.key
        return JDict(), last.key

    def resolve(self, chain: str) -> Any:
        """
        resolves the chain & returns the value

        raises an exception if the chain is invalid
        """
        value = self._prepare(chain)
        if value is None:
            raise KeyError(f"invalid chain: {chain}")
        parent, last = value
        assert isinstance(last, (int, str))
        return parent[last]

    def assertGet(self, chain: str, type_: Type[T]) -> T:
        """
        resolves the chain & returns the value

        raises an exception if the chain is invalid or the value is not the specified type
        """
        value = self._prepare(chain)
        if value is None:
            raise KeyError(f"invalid chain: {chain}")
        parent, last = value
        assert isinstance(last, (int, str))
        return parent.assertGet(last, type_)

    def optionalGet(self, chain: str, type_: Type[T]) -> Optional[T]: # pylint: disable=arguments-renamed
        """
        tries to resolve the chain

        if any key does not exist, None is returned
        if the chain resolves but is not the specified type, None is returned

        Usage: you don't know if the chain resolves
        """
        value = self._prepare(chain, preventException = True)
        if value is None:
            return None
        obj, last = value
        assert isinstance(last, (int, str))
        return obj.optionalGet(last, type_)

    def optionalCast(self, chain: str, type_: Type[T]) -> Optional[T]: # pylint: disable=arguments-renamed
        """
        tries to resolve the chain

        if any key does not exist, None is returned
        if the chain resolves but is not the specified type, None is returned

        Usage: you don't know if the chain resolves
        """
        value = self._prepare(chain, preventException = True)
        if value is None:
            return None
        obj, last = value
        assert isinstance(last, (int, str))
        return obj.optionalCast(last, type_)

    def ensure(self, chain: str, type_: Type[T], default: Optional[T] = None) -> T: # pylint: disable=arguments-renamed
        """
        tries to resolve the chain

        if any key does not exist, the default value is returned
        if the chain resolves but is not the specified type, the default value is returned

        Usage: you don't know if the chain resolves
        """
        value = self._prepare(chain, preventException = True)
        if value is None:
            return default or type_()
        obj, last = value
        assert isinstance(last, (int, str))
        return obj.ensure(last, type_, default)

    def ensureCast(self, chain: str, type_: Type[T], default: Optional[T] = None) -> T: # pylint: disable=arguments-renamed
        """
        tries to resolve the chain

        if any key does not exist, the default value is returned
        if the chain resolves but is not the specified type, the default value is returned

        Usage: you don't know if the chain resolves
        """
        value = self._prepare(chain, preventException = True)
        if value is None:
            return default or type_()
        obj, last = value
        assert isinstance(last, (int, str))
        return obj.ensureCast(last, type_, default)



# You're dealing with nested objects

from typing import Optional
from pyaddict import JDict

match = {
    "home": {
        "name": "Liverpool",
        "score": 2
    },
    "away": {
        "name": "Chelsea",
        "score": 1
    }
}

match2 = {
    "home": {
        "name": "Manchester City",
        "score": 3
    },
    # opponent is not known yet
}

jdict = JDict(match)
jdict2 = JDict(match2)

# with normal dicts
print(match["home"]) # { "name": "Liverpool", "score": 2 }
print(match["away"]) # { "name": "Chelsea", "score": 1 }
print(match2["home"]) # { "name": "Manchester City", "score": 3 }
try:
    print(match2["away"]) # KeyError: 'away'
except KeyError as e:
    print(e)

# JDict can handle this
print(jdict.ensure("home", dict)) # { "name": "Liverpool", "score": 2 }
print(jdict.ensure("away", dict)) # { "name": "Chelsea", "score": 1 }
print(jdict2.ensure("home", dict)) # { "name": "Manchester City", "score": 3 }
print(jdict2.ensure("away", dict)) # { }

# if you're now to access the name of the team, you would have to do this
print(match["home"]["name"]) # "Liverpool"
print(match["away"]["name"]) # "Chelsea"
print(match2["home"]["name"]) # "Manchester City"
try:
    print(match2["away"]["name"]) # KeyError: 'away'
except KeyError as e:
    # here you can't even access the name of the team
    print(e)
# you could fix the key error:
print(match2.get("away", dict()).get("name", str())) # ""

# JDict can handle this
print(jdict2.ensureCast("away", JDict).ensure("name", str)) # ""

# so is it just more verbose?
# no, because jdict offers more features
print(type(match2.get("home", dict()).get("name"))) # Any | None -> <class 'str'>
print(type(match2.get("away", dict()).get("name"))) # Any | None -> <class 'NoneType'>
print(type(match2.get("away", dict()).get("name", None))) # Any | None -> <class 'NoneType'>
# you can't specify the type you want; you can only specify the default value
# if you want to specify the type, you have to do this
name: Optional[str] = match2.get("away", dict()).get("name")
# but this is not safe, because you can't be sure that the value is a string:
score: Optional[str] = match2.get("home", dict()).get("score")
print(type(score), score) # <class 'int'> 3

# JDict offers this out of the box
print(type(jdict2.ensureCast("home", JDict).optionalGet("name", str))) # str | None -> <class 'str'>
print(type(jdict2.ensureCast("away", JDict).optionalGet("name", str))) # str | None -> <class 'NoneType'>
# your IDE knows the type of the value already
name = jdict2.ensureCast("away", JDict).optionalGet("name", str)
score = jdict2.ensureCast("home", JDict).optionalGet("score", str)
print(type(score), score) # <class 'NoneType'> None
score = jdict2.ensureCast("home", JDict).optionalGet("score", int)
print(type(score), score) # <class 'int'> 3



# You're dealing with nested lists

from pyaddict import JList

array = [ [ 1, 2, 3 ], [ "4", "5", "6" ], "I'm not even a list" ]
jlist = JList(array)

# with normal lists
print(array[0]) # [ 1, 2, 3 ]
print(array[1]) # [ "4", "5", "6" ]
print(array[2]) # "I'm not even a list"
# easy until now
print(array[0][0]) # 1
print(array[1][0]) # "4"
print(array[2][0]) # I
# but this might not be what you want

# JList can handle this
print(jlist.ensure(0, list)) # [ 1, 2, 3 ]
print(jlist.ensure(1, list)) # [ "4", "5", "6" ]
print(jlist.ensure(2, list)) # [ ]
print(jlist.ensure(2, str)) # "I'm not even a list"
print(jlist.ensureCast(2, list)) # ['I', "'", 'm', ' ', 'n', 'o', 't', ' ', 'e', 'v', 'e', 'n', ' ', 'a', ' ', 'l', 'i', 's', 't']

print(jlist.ensureCast(0, JList).ensure(0, int)) # 1
print(jlist.ensureCast(1, JList).ensure(0, str)) # "4"
print(jlist.ensureCast(1, JList).ensureCast(0, int)) # 4
print(jlist.ensureCast(2, JList).ensure(0, str)) # I



# -*- coding: utf-8 -*-
"""pyaddict"""
__copyright__ = ("Copyright (c) 2022 https://github.com/dxstiny")

import unittest

from pyaddict import JDict
from pyaddict.pyaddict import JList, JListIterator
from pyaddict.schema import  Object, String, Integer, Float, Boolean, Array
from pyaddict.types import JObject


dictionary: JObject = {
    "stringvalue": "value1",
    "intvalue": 1,
    "floatvalue": 1.0,
    "boolvalue": True,
    "listvalue": [1, 2, 3],
    "intAsString": "5",
    "dictvalue": {
        "stringvalue": "value2",
        "intvalue": 2,
        "floatvalue": 2.0,
        "boolvalue": True,
        "listvalue": [2, 3, 4],
        "dictlist": [
            {
                "stringvalue": "value3",
                "intvalue": 3,
                "floatvalue": 3.0,
                "boolvalue": True,
            },
            {
                "stringvalue": "value4",
                "intvalue": 4,
                "floatvalue": 4.0,
                "boolvalue": True,
            },
        ],
    }
}

jdict = JDict(dictionary)


class TestAllDict(unittest.TestCase):
    def test_assertGet(self) -> None:
        self.assertEqual(jdict.assertGet("stringvalue", str), "value1")
        self.assertEqual(jdict.assertGet("intvalue", int), 1)
        self.assertEqual(jdict.assertGet("floatvalue", float), 1.0)
        self.assertEqual(jdict.assertGet("boolvalue", bool), True)
        self.assertEqual(jdict.assertGet("listvalue", list), [1, 2, 3])

    def test_optionalCast(self) -> None:
        self.assertEqual(jdict.optionalCast("stringvalue", str), "value1")
        self.assertEqual(jdict.optionalCast("intvalue", int), 1)
        self.assertEqual(jdict.optionalCast("floatvalue", float), 1.0)
        self.assertEqual(jdict.optionalCast("boolvalue", bool), True)
        self.assertEqual(jdict.optionalCast("listvalue", list), [1, 2, 3])

    def test_ensure(self) -> None:
        self.assertEqual(jdict.ensure("stringvalue", str), "value1")
        self.assertEqual(jdict.ensure("intvalue", int), 1)
        self.assertEqual(jdict.ensure("floatvalue", float), 1.0)
        self.assertEqual(jdict.ensure("boolvalue", bool), True)
        self.assertEqual(jdict.ensure("listvalue", list), [1, 2, 3])

    def test_ensureCast(self) -> None:
        self.assertEqual(jdict.ensureCast("stringvalue", str), "value1")
        self.assertEqual(jdict.ensureCast("intvalue", int), 1)
        self.assertEqual(jdict.ensureCast("floatvalue", float), 1.0)
        self.assertEqual(jdict.ensureCast("boolvalue", bool), True)
        self.assertEqual(jdict.ensureCast("listvalue", list), [1, 2, 3])

    def test_nested(self) -> None:
        self.assertEqual(jdict.ensureCast("dictvalue", JDict)
                                    .assertGet("stringvalue", str),
                         "value2")
        self.assertEqual(jdict.ensureCast("dictvalue", JDict)
                                    .assertGet("intvalue", int),
                         2)
        self.assertEqual(jdict.ensureCast("dictvalue", JDict)
                                    .assertGet("floatvalue", float),
                         2.0)
        self.assertEqual(jdict.ensureCast("dictvalue", JDict)
                                    .assertGet("boolvalue", bool),
                         True)
        self.assertEqual(jdict.ensureCast("dictvalue", JDict)
                                    .assertGet("listvalue", list),
                         [2, 3, 4])


class TestUseCases(unittest.TestCase):
    def test_usecase_dict(self) -> None:
        self.assertIsInstance(jdict.ensure("intAsString", int), int)
        self.assertEqual(jdict.ensure("intAsString", int), 0)

        self.assertIsInstance(jdict.ensureCast("intAsString", int), int)
        self.assertEqual(jdict.ensureCast("intAsString", int), 5)

        self.assertRaises(AssertionError, jdict.assertGet, "intAsString", int)

        self.assertEqual(jdict.optionalCast("intAsString", int), 5)

        self.assertIsNone(jdict.optionalGet("intAsString", int))

    def test_usecase_list(self) -> None:
        jlist = jdict.ensureCast("listvalue", JList)
        self.assertEqual(len(jlist), 3)

        self.assertEqual(jlist.assertGet(0, int), 1)
        self.assertEqual(jlist.assertGet(1, int), 2)
        self.assertEqual(jlist.assertGet(2, int), 3)
        self.assertRaises(AssertionError, jlist.assertGet, 3, int)

        self.assertEqual(jlist.optionalCast(0, int), 1)
        self.assertEqual(jlist.optionalCast(1, int), 2)
        self.assertEqual(jlist.optionalCast(2, int), 3)
        self.assertIsNone(jlist.optionalCast(3, int))

        self.assertEqual(jlist.ensure(0, int), 1)
        self.assertEqual(jlist.ensure(1, int), 2)
        self.assertEqual(jlist.ensure(2, int), 3)
        self.assertEqual(jlist.ensure(3, int), 0)

        self.assertEqual(jlist.ensureCast(0, str), "1")
        self.assertEqual(jlist.ensureCast(1, str), "2")
        self.assertEqual(jlist.ensureCast(2, str), "3")
        self.assertEqual(jlist.ensureCast(3, str), "")

        self.assertIsInstance(jlist.iterator(), JListIterator)
        self.assertEqual(list(jlist.iterator()), [1, 2, 3])

        self.assertEqual(jlist.iterator().ensure(int), [1, 2, 3])
        self.assertEqual(jlist.iterator().ensure(str), ["", "", ""])
        self.assertEqual(jlist.iterator().ensureCast(int), [1, 2, 3])
        self.assertEqual(jlist.iterator().ensureCast(str), ["1", "2", "3"])
        self.assertEqual(jlist.iterator().assertGet(int), [1, 2, 3])
        self.assertRaises(AssertionError, jlist.iterator().assertGet, str)
        self.assertEqual(jlist.iterator().optionalCast(int), [1, 2, 3])
        self.assertEqual(jlist.iterator().optionalCast(str), [None, None, None])
        self.assertEqual(jlist.iterator().optionalGet(int), [1, 2, 3])
        self.assertEqual(jlist.iterator().optionalGet(str), [None, None, None])

    def test_chaining(self) -> None:
        chain = jdict.chain()
        self.assertEqual(chain.ensureCast("dictvalue.listvalue", JList)
                              .iterator()
                              .ensure(int),
                         [2, 3, 4])
        self.assertEqual(chain.ensureCast("dictvalue.listvalue.[0]", int), 2)
        self.assertIsNone(chain.optionalGet("dictvalue.dictlist.[2]", dict))
        self.assertRaises(IndexError, chain.__getitem__, "dictvalue.dictlist.[2].value") # []
        self.assertIsNone(chain.optionalGet("dictvalue.dictlist.[2]?.stringvalue", str))

    def test_serialise(self) -> None:
        self.assertIsInstance(jdict.toString(), str)
        self.assertIsInstance(JDict.fromString(jdict.toString()), JDict)

        self.assertEqual(JList.fromString(jdict.toString()), [ ])

        self.assertIsInstance(JList.fromString(jdict.toString()), JList)
        self.assertIsInstance(JList.fromString(jdict.toString()).toString(), str)


class TestSchemas(unittest.TestCase):
    def test_schemas(self) -> None:
        schema = Object({
            "stringvalue": String(),
            "intvalue": Integer(),
            "floatvalue": Float(),
            "boolvalue": Boolean(),
            "listvalue": Array(Integer()),
            "intAsString": Integer().coerce(),
            "dictvalue": Object({
                "stringvalue": String(),
                "intvalue": Integer(),
                "floatvalue": Float(),
                "boolvalue": Boolean(),
                "listvalue": Array(Integer()),
                "dictlist": Array(Object({
                    "stringvalue": String(),
                    "intvalue": Integer(),
                    "floatvalue": Float(),
                    "boolvalue": Boolean(),
                }))
            })
        })
        self.assertIsInstance(schema(jdict)["intAsString"], int)
        self.assertEqual(schema.valid(jdict), True)
        self.assertEqual(schema.error(jdict), None)

    def test_schema_noadditional(self) -> None:
        schema = Object({
            "stringvalue": String()
        })
        self.assertEqual(schema.valid(jdict), False)

    def test_schema_int(self) -> None:
        schema = Integer().min(0, False).max(10, False)
        self.assertEqual(schema.valid(1.0), False)

        self.assertEqual(schema.valid(0), False)
        self.assertEqual(schema.valid(1), True)
        self.assertEqual(schema.valid(9), True)
        self.assertEqual(schema.valid(10), False)

        schema2 = Integer().min(0, True).max(10, False)
        self.assertEqual(schema2.valid(0), True)
        self.assertEqual(schema2.valid(1), True)
        self.assertEqual(schema2.valid(10), False)

    def test_schema_string(self) -> None:
        schema = String().url()
        self.assertEqual(schema.valid("http://www.google.com"), True)
        self.assertEqual(schema.valid("http://www.google.com/"), True)
        self.assertEqual(schema.valid("http://www.google.com/abc"), True)
        self.assertEqual(schema.valid("test"), False)

        schema2 = String().email()
        self.assertEqual(schema2.valid("test@mail.com"), True)
        self.assertEqual(schema2.valid("test@mail"), False)

        schema3 = String().regex(r"^[a-z]+$")
        self.assertEqual(schema3.valid("test"), True)
        self.assertEqual(schema3.valid("test123"), False)

if __name__ == '__main__':
    unittest.main()



"""json schema validation inspired by zod"""

from __future__ import annotations
from typing import Any, Dict, List, Optional, TypeVar, Pattern, Generic, Union, Type
from abc import ABC, abstractmethod
import re
import json
from enum import Enum

from pyaddict.types import JObject


T = TypeVar("T")
U = TypeVar("U")
IS = TypeVar("IS")


class ValidationState(Enum):
    """ValidationState"""
    Valid = 0
    Invalid = 1


class ValidationError(ValueError):
    """ValidationError"""
    def __init__(self,
                 message: str,
                 path: List[str],
                 validation: str) -> None:
        super().__init__(message)
        self._message = message
        self._path = path
        self._validation = validation

    def __repr__(self) -> str:
        return f"ValidationError({self._message}, {self.formattedPath})"

    def __str__(self) -> str:
        return self.__repr__()

    @property
    def formattedPath(self) -> str:
        """a friendly path for the error"""
        if len(self._path) == 0:
            return '(root): ' + self._validation
        path = [ str(p) for p in self._path ]
        return '.'.join(path) + ': ' + self._validation

    @property
    def path(self) -> List[str]:
        """the path to the error"""
        return self._path

    @property
    def message(self) -> str:
        """the error message"""
        return self._message

    @property
    def validation(self) -> str:
        """the validation that failed"""
        return self._validation

    @staticmethod
    def inherit(error: Optional[ValidationError], path: List[str]) -> ValidationError:
        """inherit a ValidationError (combines the paths)"""
        assert error is not None
        return ValidationError(error.message, path + error.path, error.validation)


class ValidationResult(Generic[T]):
    """OptionalData"""
    def __init__(self,
                 state: ValidationState,
                 data: Optional[T] = None,
                 error: Optional[ValidationError] = None) -> None:
        self._state = state
        self._data = data
        self._error = error

    def valid(self) -> bool:
        """is the value valid?"""
        return self._state == ValidationState.Valid

    def __bool__(self) -> bool:
        return self.valid()

    def __repr__(self) -> str:
        if self._state == ValidationState.Valid:
            return f"ValidationResult({self._data})"
        return f"ValidationResult({self._error})"

    def unwrap(self) -> T:
        """unwrap the value if valid, otherwise raise an error"""
        if self._state == ValidationState.Valid:
            assert self._data is not None
            return self._data
        raise ValueError("unwrap called on invalid value")

    def unwrapOr(self, default: T) -> T:
        """unwrap the value if valid, otherwise return the default"""
        if self._state == ValidationState.Valid:
            assert self._data is not None
            return self._data
        return default

    def expect(self, msg: Optional[str] = None) -> T:
        """
        unwrap the value if valid,
        otherwise raise an error with the message, if provided,
        otherwise raise the original error
        """
        if self._state == ValidationState.Valid:
            assert self._data is not None
            return self._data
        if msg is None:
            assert self._error is not None
            raise self._error # pylint: disable=raising-bad-type # (mypy bug)
        raise ValueError(msg)

    @property
    def error(self) -> Optional[ValidationError]:
        """the error if invalid"""
        return self._error

    def invalidate(self, error: Optional[ValidationError] = None) -> None:
        """invalidate the value"""
        self._state = ValidationState.Invalid
        self._error = error

    def update(self, value: ValidationResult[T]) -> None:
        """update the value"""
        if not value:
            self.invalidate(value.error)

    @staticmethod
    def ok(data: U) -> ValidationResult[U]:
        """create a valid result"""
        return ValidationResult(ValidationState.Valid, data=data)

    @staticmethod
    def err(error: Optional[ValidationError] = None) -> ValidationResult[U]:
        """create an invalid result"""
        return ValidationResult(ValidationState.Invalid, error=error)


class ISchemaType(ABC, Generic[IS]):
    """Schema Type"""
    def __init__(self) -> None:
        super().__init__()
        self._min: Optional[int] = None
        self._max: Optional[int] = None
        self._optional: bool = False
        self._coerce: bool = False

    def coerce(self) -> IS:
        """coerce the value (float -> int, str -> bool, etc.)"""
        self._coerce = True
        return self # type: ignore

    def optional(self) -> IS:
        """make the value optional"""
        self._optional = True
        return self # type: ignore

    @abstractmethod
    def validate(self, value: T) -> ValidationResult[T]:
        """
        validates the value

        returns a ValidationResult,
        which contains an error if the value is invalid
        or the value if valid
        """
        return ValidationResult.ok(value)

    def _coerceValue(self, value: Any, to: Type[T]) -> ValidationResult[T]:
        """coerce the value"""
        if self._coerce:
            # str -> bool
            if to == bool and isinstance(value, str):
                if value.lower() in ("true", "1", "yes", "y"):
                    return ValidationResult.ok( True ) # type: ignore
                if value.lower() in ("false", "0", "no", "n"):
                    return ValidationResult.ok( False ) # type: ignore
                return ValidationResult.err(
                    ValidationError(f"{value} is not a boolean",
                                    [],
                                    "coerce"))

            # str -> dict
            if to in (dict, list) and isinstance(value, str): # type: ignore # (comparison-overlap)
                return ValidationResult.ok( json.loads(value) )

            # * -> * (hope & pray)
            return ValidationResult.ok( to(value) ) # type: ignore
        if not isinstance(value, to):
            return ValidationResult.err( ValidationError(f"{value} is not of type {to}",
                                                         [], "coerce") )
        return ValidationResult.ok( value )

    def __call__(self, value: T) -> T:
        """
        expect the value to be valid,
        otherwise raise an error
        """
        return self.expect(value)

    def expect(self, value: T, msg: Optional[str] = None) -> T:
        """
        expect the value to be valid,
        otherwise raise an error with the message, if provided,
        otherwise raise the original error
        """
        return self.validate(value).expect(msg)

    def error(self, value: T) -> Optional[ValidationError]:
        """
        get the error if the value is invalid,
        otherwise return None
        """
        return self.validate(value).error

    def valid(self, value: T) -> bool:
        """
        is the value valid?
        """
        return self.validate(value).valid()


class IWithEnum(ABC, Generic[T, IS]):
    """Schema Type with Enum"""
    def __init__(self) -> None:
        super().__init__()
        self._enum: Optional[List[T]] = None

    def validate(self, value: T) -> ValidationResult[T]:
        """validate the value"""
        if self._enum is not None and value not in self._enum:
            return ValidationResult.err(
                ValidationError(f"{value} is not in {self._enum}",
                                [],
                                "enum"))
        return ValidationResult.ok(value)

    def enum(self, *values: U) -> IS:
        """set the enum values"""
        self._enum = list(values)
        return self # type: ignore


class IWithLength(ABC, Generic[T, IS]):
    """Schema Type with Length (array, string, etc.)"""
    def __init__(self) -> None:
        super().__init__()
        self._min: Optional[int] = None
        self._max: Optional[int] = None
        self._minInclusive: bool = False
        self._maxInclusive: bool = False

    @abstractmethod
    def length(self, value: Any) -> int:
        """get the length of the value"""

    def validate(self, value: T) -> ValidationResult[T]:
        """validate the value"""
        length = self.length(value)
        if self._min is not None:
            if self._minInclusive:
                if length < self._min:
                    return ValidationResult.err(
                        ValidationError(f"expected {self.length(value)} to be greater than or equal to {self._min}", # pylint: disable=line-too-long
                                        [],
                                        "min"))
            else:
                if length <= self._min:
                    return ValidationResult.err(
                        ValidationError(f"expected {self.length(value)} to be greater than {self._min}", # pylint: disable=line-too-long
                                        [],
                                        "min"))
        if self._max is not None:
            if self._maxInclusive:
                if length > self._max:
                    return ValidationResult.err(
                        ValidationError(f"expected {self.length(value)} to be less than or equal to {self._max}", # pylint: disable=line-too-long
                                        [],
                                        "max"))
            else:
                if length >= self._max:
                    return ValidationResult.err(
                        ValidationError(f"expected {self.length(value)} to be less than {self._max}", # pylint: disable=line-too-long
                                        [],
                                        "max"))
        return ValidationResult.ok(value)

    def min(self, min_: int, inclusive: bool = True) -> IS:
        """set the min length"""
        self._min = min_
        self._minInclusive = inclusive
        return self # type: ignore

    def max(self, max_: int, inclusive: bool = True) -> IS:
        """set the max length"""
        self._max = max_
        self._maxInclusive = inclusive
        return self # type: ignore


class String(ISchemaType["String"], IWithLength[str, "String"], IWithEnum[str, "String"]):
    """String Schema Type"""
    def __init__(self) -> None:
        super().__init__()
        self._regex: Optional[Union[Pattern[str], str]] = None
        self._regexType: Optional[str] = None
        self._regexMessage: Optional[str] = None

    def length(self, value: str) -> int:
        return len(value)

    def validate(self, value: Any) -> ValidationResult[str]:
        result = self._coerceValue(value, str)

        if not result:
            return result

        result.update(IWithEnum.validate(self, result.unwrap()))
        if not result:
            return result

        result.update(IWithLength.validate(self, result.unwrap()))
        if not result:
            return result

        if self._regex is not None:
            if isinstance(self._regex, str):
                self._regex = re.compile(self._regex)
            if not self._regex.match(result.unwrap()):
                assert self._regexType is not None and self._regexMessage is not None
                result.invalidate(
                    ValidationError(self._regexMessage, [], self._regexType))
        return result

    def regex(self, regex: Union[str, Pattern[str]]) -> String:
        """validate the value against a regex"""
        self._regex = regex
        self._regexType = "regex"
        self._regexMessage = f"string didn't match {self._regex}"
        return self

    def email(self) -> String:
        """validate the value as an email"""
        self._regex = r"^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$"
        self._regexType = "email"
        self._regexMessage = "string is not a valid email"
        return self

    def url(self) -> String:
        """validate the value as a url"""
        self._regex = r"^(https?|ftp)://[^\s/$.?#].[^\s]*$"
        self._regexType = "url"
        self._regexMessage = "string is not a valid url"
        return self


class Integer(ISchemaType["Integer"], IWithEnum[int, "Integer"], IWithLength[int, "Integer"]):
    """Integer Schema Type"""
    def validate(self, value: Any) -> ValidationResult[int]:
        result = self._coerceValue(value, int)
        if not result:
            return result

        result.update(IWithEnum.validate(self, result.unwrap()))
        if not result:
            return result

        result.update(IWithLength.validate(self, result.unwrap()))
        return result

    def length(self, value: int) -> int:
        return value


class Float(ISchemaType["Float"], IWithEnum[float, "Float"], IWithLength[float, "Float"]):
    """Float Schema Type"""
    def validate(self, value: Any) -> ValidationResult[float]:
        result = self._coerceValue(value, float)
        if not result:
            return result

        result.update(IWithEnum.validate(self, value))
        if not result:
            return result

        result.update(IWithLength.validate(self, value))
        return result

    def length(self, value: float) -> float:
        return value


class Boolean(ISchemaType["Boolean"], IWithEnum[bool, "Boolean"]):
    """Boolean Schema Type"""
    def validate(self, value: Any) -> ValidationResult[bool]:
        result = self._coerceValue(value, bool)
        if not result:
            return result
        result.update(IWithEnum.validate(self, value))
        return result


class Object(ISchemaType["Object"]):
    """Object Schema Type"""
    def __init__(self,
                 body: Optional[Dict[str, ISchemaType[Any]]] = None,
                 additionalProperties: bool = False) -> None:
        super().__init__()
        self._body = body
        self._allowAdditionalProperties = additionalProperties

    def withAdditionalProperties(self, withAdditionalProperties: bool = True) -> Object:
        """disallow additional properties"""
        self._allowAdditionalProperties = withAdditionalProperties
        return self

    def validate(self, value: Any) -> ValidationResult[JObject]:
        result = self._coerceValue(value, dict)
        if not result:
            return result

        body = self._body or {}

        for key, schema in body.items():
            if key not in result.unwrap():
                if not schema._optional: # pylint: disable=protected-access
                    result.invalidate(
                        ValidationError(f"expected {key} to be present",
                                        [],
                                        "required"))
                    return result
                continue
            keyRes = schema.validate(value[key])
            if not keyRes:
                result.invalidate(ValidationError.inherit(keyRes.error, [key]))
                return result
            value[key] = keyRes.unwrap()

        if not self._allowAdditionalProperties:
            for key in result.unwrap():
                if key not in body:
                    result.invalidate(ValidationError(f"unexpected property {key}",
                                                     [],
                                                     "additionalProperties"))
                    return result

        return result

class Array(ISchemaType["Array"], IWithLength[List[Any], "Array"]):
    """Array Schema Type"""
    def __init__(self, item: ISchemaType[Any]) -> None:
        super().__init__()
        self._item = item

    def validate(self, value: Any) -> ValidationResult[List[Any]]:
        result = self._coerceValue(value, list)
        if not result:
            return result

        result.update(IWithLength.validate(self, result.unwrap()))
        if not result:
            return result

        values: List[Any] = [ ]
        for item in result.unwrap():
            res = self._item.validate(item)
            if not res.valid():
                result.invalidate(ValidationError.inherit(res.error, ["[]"]))
                return result
            values.append(res.unwrap())
        return ValidationResult.ok(values)

    def length(self, value: List[Any]) -> int:
        return len(value)


class OneOf(ISchemaType["OneOf"]):
    """OneOf Schema Type"""
    def __init__(self, *items: ISchemaType[Any]) -> None:
        super().__init__()
        self._items = items

    def validate(self, value: Any) -> ValidationResult[Any]:
        errors: List[ValidationError] = [ ]
        for item in self._items:
            res = item.validate(value)
            if res.valid(): # use typeguard
                return res
            assert res.error
            errors.append(res.error)
        return ValidationResult.err(
            ValidationError("value didn't match any of the schemas", [], "oneOf"))
